import os
import os.path
import glob
import json
import uuid
import datetime
import shutil
import sys
import traceback
import copy
from urllib.parse import urlparse
import threading
import time
import tempfile
import zipfile
import fnmatch
import base64
import socket
import re
import feedparser

#############
from collections import OrderedDict
from viper import github
import viper.utils as vutils
#############

import ide.viper
import viper.viperconf as vconf
import viper.viperlogger
import viper.toolbelt as tb
from ide.vipergit import ViperGit
from ide.viperthreads import ViperThread
import compiler.compiler as compiler
import uplinker.uplinker as uplinker

logger = viper.viperlogger.logger(__name__,"ide")


class ViperControl():
    def __init__(self):
        self.client = ide.viper.getClient()
        self.hwlistener=ide.viper.getHwListener()
        self.git = ViperGit()
        self.projdir = ide.viper.projdir
        self.consoles = {}
        self.devices = {}
        self.dev_props = {}
        self.projects = {}
        self.bcodes = {}
        self.db = ide.viper.getDB()
        self.backend = ide.viper.getBackend()
        ## LOAD PROJECTS
        self.projects = self.db.get_projects()
        #sanity check
        tprj = copy.deepcopy(self.projects)
        for uid,proj in self.projects.items():
            if not os.path.exists(self.dname(proj)):
                del tprj[proj["id"]]
            else:
                for fname in proj["files"]:
                    if not os.path.exists(self.fname(proj,fname)):
                        del tprj[proj["id"]]["files"][fname]
        self.projects = tprj

        #clean leftovers projects
        prjdirs = next(os.walk(self.projdir))[1]
        for pdir in prjdirs:
            if not pdir in self.projects:
                shutil.rmtree(os.path.join(self.projdir,pdir),ignore_errors=True)

        ## LOAD DEVICE PROPS
        self.dev_props = self.db.get_devices()

        self.vms = {}


        ##send last project
        self.send_last_project()
        self.vpm = tb.vpm
        self.send_vpm()
        self.trepo = True
        self.rrepo = True

        tb.load_modules_info()

    def open_project(self,msg):
        data = msg["data"]
        if data["id"] in self.projects:
            res = self.projects[data["id"]]
            self.client.answer(msg,"ok")
            self.send_project(res)
            return
        self.client.answer(msg,"ko","Project does not exists!")

    def save_project(self,msg):
        data = msg["data"]

        if data["id"] in self.projects:
            res = self.projects[data["id"]]
            s1 = set(res["files"])
            s2 = set(data["files"])
            no_need_update = s1<=s2 and s1>=s2
            res["files"]=data["files"]
            res["synched"]="ahead"
            self._save_project(res)
            self.projects[data["id"]]=res
            self.store()
            self.client.answer(msg,"ok")
            if not no_need_update:
                self.send_project(res,False)
            return
        self.client.answer(msg,"ko","Project does not exists!")

    def _new_project_dir(self,proj):
        uid = str(uuid.uuid4())
        logger.info("new project uid %s",uid)
        while os.path.exists(os.path.join(self.projdir,uid)):
            uid = uuid.uuid4()
        proj["id"] = uid
        os.mkdir(self.dname(proj))
        return uid

    def _new_project_git(self,proj,msg):
        try:
            self.git.new_repo(self.dname(proj))
            self.git.do_commit(self.dname(proj),"First_Commit")
        except ide.vipergit.GitException as ge:
            logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
            shutil.rmtree(self.dname(proj),ignore_errors=True)
            self.client.answer(msg,"ko","Something wrong with git: "+str(ge))
            ide.viper.say("PROJECT","oops, git failed :(")
            return False

        if self.backend.logged:
            try:
                res= self.backend.new_project(proj)
                if res:
                    proj["remoteId"]=res["remoteId"]
                    self.git.add_remote(ide.viper.viper_origin,self.backend.get_project_repo_url(proj),self.dname(proj))
                    self.git.first_push(self.dname(proj),ide.viper.viper_origin)
                    ide.viper.say("PROJECT","Project "+proj["name"]+" is in the cloud")
                else:
                    ide.viper.say("PROJECT","Project "+proj["name"]+" can't be saved to cloud")
            except ide.vipergit.GitException as ge:
                logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
                ide.viper.say("PROJECT","Project "+proj["name"]+" can't be saved to cloud: "+str(ge))
        else:
            ide.viper.say("PROJECT","Project "+proj["name"]+" can't be sync'd, you are not logged")
        return True

    def new_project(self,msg):
        data = msg["data"]
        proj = self._project_exists(data["name"])
        ide.viper.say("PROJECT","Creating new project "+data["name"])
        if proj:
            self.client.answer(msg,"ko","Project already exists!")
            self.send_project(proj)
            ide.viper.say("PROJECT","oops, that project already exists")
            return False

        # check for same namespace.path libs
        if "namespace" in data and "mod_name" in data:
            for p in self.projects:
                if "mod_name" in self.projects[p] and self.projects[p]["mod_name"]==data["mod_name"] and self.projects[p]["namespace"]==data["namespace"]:
                    self.client.answer(msg,"ko","Library path ["+data["mod_name"]+"] already exists!")
                    ide.viper.say("PROJECT","oops, that library name already exists")
                    return False
        proj = data
        uid = self._new_project_dir(proj)
        proj["synched"]="ahead"
        if 'namespace' in data:
            proj.update({"mode": 'release'})
            # proj["files"]={"readme.md":self.generate_project_md(proj),"main.py":self.generate_library_main_py(proj)}
            proj["files"]=OrderedDict({"readme.md":self.generate_project_md(proj),"main.py":self.generate_library_main_py(proj)})
            proj["files"].update({ 'package.json': ''})
            proj["files"].update({ proj["mod_name"] + '.py': '', 'examples/': '', 'docs/' : ''})
            proj["files"].update({ 'docs/docs.json': self.generate_library_docs_json(proj), 'docs/index.rst': self.generate_library_docs_rst(proj)})
        else:
            proj["files"]={"project.md":self.generate_project_md(proj),"main.py":self.generate_main_py(proj)}
        ide.viper.say("PROJECT","Creating git repo for "+data["name"])

        self._save_project(proj)

        if not self._new_project_git(proj,msg):
            return False

        self.projects[uid]=proj
        self.store()
        self.client.answer(msg,"ok")
        self.send_project(proj)
        self.send_projects()
        return True

    def list_projects(self,msg):
        self.send_projects()
        self.client.answer(msg,"ok")

    def check_projects_sync(self,msg):
        if not self.backend.logged:
            self.client.answer(msg,"ko","You are not logged!")
            return
        for pid,proj in self.projects.items():
            proj["synched"]="unknown"
            try:
                self.git.do_commit(self.dname(proj),"Sync_Commit")
                proj["synched"] = self.git.check_status(self.dname(proj),ide.viper.viper_origin)
            except ide.vipergit.GitException as ge:
                logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
            #print("PROJECT",proj["name"],proj["synched"])
        self.send_projects()
        self.client.answer(msg,"ok")


    def update_project(self,msg):
        data = msg["data"]
        if data["id"] in self.projects:
            proj = self.projects[data["id"]]
            if data["remove"]:
                logger.info("removing %s",self.dname(proj))
                shutil.rmtree(self.dname(proj),ignore_errors=True)
                self.db.remove_project(data["id"])
                ide.viper.say("PROJECT",proj["name"]+ " deleted!")
                del self.projects[data["id"]]
                self.store()
                if "remoteId" in proj:
                    self.backend.delete_project(proj)
                self.client.answer(msg,"ok","Project deleted!")
                self.send_projects()
            else:
                prj =self._project_exists(data["name"])
                if prj and prj["id"]!=proj["id"]:
                    self.client.answer(msg,"ko","Project already exists!")
                    return
                proj["name"]=data["name"]
                proj["desc"]=data["desc"]
                proj["tags"]=data["tags"]
                proj["repos"]=data["repos"]
                self.store()
                #self.send_project(proj)
                if "remoteId" in proj:
                    self.backend.update_project(proj)
                ide.viper.say("PROJECT",proj["name"]+ " updated!")
                self.client.answer(msg,"ok","Project updated!")
                self.send_projects()
            return
        self.client.answer(msg,"ko","Projects does not exists!")

    def fork_project(self,msg):
        data = msg["data"]
        proj = self._project_exists(data["name"])
        if proj:
            self.client.answer(msg,"ko","Project already exists!")
            return False
        proj = data
        repo = data["repo"]
        del proj["repo"]
        uid = self._new_project_dir(proj)
        if not repo["url"].startswith("https://"):
            #local url...
            #print("FORK: ",repo)
            proj["repos"]=[]
            shutil.rmtree(self.dname(proj),ignore_errors=True)
            if repo["url"] in self.projects:
                #cloning of a local project
                shutil.copytree(os.path.join(self.projdir,repo["url"]),self.dname(proj))
                shutil.rmtree(os.path.join(self.dname(proj),".git"),ignore_errors=True)
            else:
                shutil.copytree(repo["url"],self.dname(proj))
            if not self._new_project_git(proj,msg):
                self.client.answer(msg,"ko","Something wrong with project cloning")
                return False
        else:
            #remote repo
            try:
                self.git.clone(self.dname(proj),self.git.origin(repo["url"],repo["user"],repo["pass"]))
            except ide.vipergit.GitException as ge:
                logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
                shutil.rmtree(self.dname(proj),ignore_errors=True)
                self.client.answer(msg,"ko","Something wrong with git: "+str(ge))
                return
            proj["repos"]=[repo]
        proj["synched"]="ahead"
        self.fill_from_path(proj)
        self.projects[uid]=proj
        self.send_project(proj)
        self.send_projects()
        self.store()
        self.client.answer(msg,"ok")


    def file_import(self,msg):
        data = msg["data"]
        if data["id"] not in self.projects:
            self.client.answer(msg,"ko","No such project")
            return
        proj = self.projects[data["id"]]
        fim = data["url"]
        if fim:
            o = urlparse(fim)
            if o.path.endswith(".zip"):
                try:
                    self._get_and_extract_archive(fim,self.dname(proj))
                    self.fill_from_path(proj)
                    self.store()
                    self.client.answer(msg,"ok")
                    self.send_project(proj)
                except Exception as e:
                    logger.error("Something wrong with file_import: %s",ide.viper.pretty_exc(e))
                    self.client.answer(msg,"ko","Error "+ide.viper.pretty_exc(e))
                return
            self.client.answer(msg,"ko","Can import only zip files!")
        else:
            for fn in data["files"]:
                with open(self.fname(proj,fn),"wb") as ff:
                    ff.write(data["files"][fn]) # it's already base64 encoded
            self.fill_from_path(proj)
            self.store()
            self.client.answer(msg,"ok")
            self.send_project(proj)

    def open_console(self,msg):
        data = msg["data"]
        if data["device"] in self.consoles and data["close"]:
            # close opened console
            self.consoles[data["device"]].stop()
            self.consoles[data["device"]].has_close.wait()
            del self.consoles[data["device"]]
            self.client.answer(msg,"ok")
            return
        elif data["device"] not in self.consoles and not data["close"]:
            # open new console
            if data["device"].startswith("dummy"):
                self.consoles[data["device"]]=FakeConsole(data["device"])
            else:
                dev = self.devices[data["device"]]
                board = self._get_board(dev)
                if not board:
                    self.client.answer(msg,"ko","No Such Board")
                    return
                if hasattr(board,"reset_before_console"):
                    board.reset()
                self.consoles[data["device"]]=ViperConsole(data["device"],board)
            self.client.answer(msg,"ok")
            return
        self.client.answer(msg,"ko","Wrong console message: "+str(self.consoles))

    def send_console(self,msg):
        data = msg["data"]
        if data["device"] in self.consoles:
            self.consoles[data["device"]].send(data["chrs"])
        self.client.answer(msg,"ok")

    def viperize(self,msg):
        data = msg["data"]
        if data["device"] not in self.devices:
                self.client.answer(msg,"ko","No Such Device")
        dev = self.devices[data["device"]]
        board = self._get_board(dev)
        if not board:
            self.client.answer(msg,"ko","No Such Device")
        if data["device"] in self.consoles:
            self.consoles[data["device"]].stop()
            self.consoles[data["device"]].has_close.wait()
            del self.consoles[data["device"]]
            self.say_close_console(data["device"])

        if data["restore"]:
            #restore fw
            ide.viper.say("VIPERIZE","Restoring original firmware...")
            res = board.restore()
            if res:
                self.client.answer(msg,"ko","Error while restoring original firmware: "+res)
                ide.viper.say("VIPERIZE","Something wrong while restoring original firmware..."+res)
                return
            #dev["viperized"]=vm["version"]
            ide.viper.say("VIPERIZE","Firmware restored!!")
            self.client.answer(msg,"ok")
        else:
            #viperize!
            if data["vm"] not in self.vms:
                self.client.answer(msg,"ko","No Such VM")
            vm = self.vms[data["vm"]]
            ide.viper.say("VIPERIZE","Viperizing...")
            res = board.burn(base64.b64decode(bytes(vm["bin"],"utf-8")))
            if res:
                self.client.answer(msg,"ko","Error while viperizing: "+res)
                ide.viper.say("VIPERIZE","Something wrong while viperizing..."+res)
                return
            #dev["viperized"]=vm["version"]
            ide.viper.say("VIPERIZE","Enjoy your viperized board!")
            self.client.answer(msg,"ok")

    def device_props(self,msg):
        data = msg["data"]
        if data["device"] not in self.dev_props:
            self.dev_props[data["device"]]={"uid":data["device"]}
        self.dev_props[data["device"]]["name"]=data["name"]
        self.dev_props[data["device"]]["desc"]=data["desc"]
        self.store()
        self.client.answer(msg,"ok")
        self.send_devices()

    def login(self,msg):
        #print("##################1")
        #TODO implement login
        email,passwd,name = self.db.get_user()
        data =msg["data"]
        if "reset" in data:
            if not self.backend.logged:
                res,txt = self.backend.reset_password(data["email"])
                if res:
                    self.client.answer(msg,"ok","Check your mail to reset the password")
                else:
                    self.client.answer(msg,"ko",txt)
                    ide.viper.say("LOGIN","Something wrong while resetting password "+txt)
            else:
                self.client.answer(msg,"ko","You must be not logged!")
        elif "signup" in data:
            if email==None:
                res,txt = self.backend.create_user(data["user"],data["password"],data["email"],data["job"])
                if res:
                    self.client.answer(msg,"ok","Check your mail to validate your account")
                else:
                    self.client.answer(msg,"ko",txt)
                    ide.viper.say("LOGIN","Something wrong while creating user "+txt)
            else:
                self.client.answer(msg,"ko","You must be not logged!")
        elif "logout" in data:
            if self.backend.logged:
                self.backend.logout()
                self.client.answer(msg,"ok")
            else:
                self.client.answer(msg,"ko","You must be logged!")
                ide.viper.say("LOGIN","Something wrong while logging out "+txt)
        else:
            if not self.backend.logged:
                res,txt = self.backend.login(None,data["password"],data["email"])
                if res:
                    self.client.answer(msg,"ok","Logged!")
                else:
                    self.client.answer(msg,"ko",txt)
                    ide.viper.say("LOGIN","Something wrong while logging in: "+txt)
            else:
                self.client.answer(msg,"ko","You must be not logged!")
        self.send_status()


    def _get_libraries_paths(self):
        res = {}
        for pid,proj in self.projects.items():
            if "namespace" in proj:
                if proj["namespace"]=="local":
                    tag="local."
                else:
                    tag="community."+proj["namespace"]
                if proj["mode"]=="testing":
                    tag="testing."+tag
                res[tag+proj["mod_name"]]=self.dname(proj)
        return res

    def _find_imported_packages(self,proj,fname):
        c = compiler.Compiler(self.fname(proj,fname),syspath=tb.syspath,prvmods=self._get_libraries_paths())
        moduletable, notfound = c.find_imports()
        print("@@@",moduletable,notfound)
        has_unknown,pkglist =self.vpm.get_packages_from_files(moduletable)
        res = { 'ok': {}, 'fail': [] }
        for k,plist in pkglist.items():
            if plist:
                for pkg in plist:
                    # print(k,"=>",pkg.fullname,pkg.installed)
                    if proj['mode'] == 'release':
                        if str(pkg.installed)[0] == 'r':
                            res['ok'].update({ pkg.fullname: str(pkg.installed) })
                    else:
                        res['ok'].update({ pkg.fullname: str(pkg.installed) })
            else:
                # file k in module moduletable[k] has no published pkg
                # check if it is a project file
                if any([ self.fname(proj,p_file) == k for p_file in proj['files']]):
                    # file k is part of the lib
                    pass
                else:
                    if not k in res['fail']:
                        res['fail'].append(moduletable[k])
        return res

    def _compile_proj(self,proj,board):
        try:
            if not board:
                board = compiler.noboard
            logger.info("Compiling proj %s for board %s",proj["id"],board.board)
            c = compiler.Compiler(self.fname(proj,"main.py"),board,self.say,syspath=tb.syspath,prvmods=self._get_libraries_paths())
            bcode,_,cobjs = c.compile()
            self.bcodes[proj["id"]]=bcode
        except compiler.CError as ce:
            errs = [{"file":self.stripfname(proj,ce.filename),"line":ce.line,"msg":ce.toCleanStr()}]
            ide.viper.getStatListener().send_stat("compile",{"errors":errs,"remoteId":proj.get("remoteId",""),"board":board.board})
            raise ce
        except Exception as e:
            logger.error("COMPILER: %s",ide.viper.pretty_exc(e))
            ide.viper.getStatListener().send_stat("compile",{"exception":ide.viper.pretty_exc(e),"remoteId":proj.get("remoteId",""),"board":board.board})
            raise e
        ide.viper.getStatListener().send_stat("compile",{"info":bcode["stats"], "board":board.board})
        return bcode,cobjs

    def compile(self,msg):
        data = msg["data"]
        if data["id"] not in self.projects:
            self.client.answer(msg,"ko","No such Project!")
            return
        proj = self.projects[data["id"]]

        if "device" in data and (not data["device"]or data["device"].startswith("-")):
            #compile for noboard
            board = None
        else:
            if data["device"] not in self.devices :
                self.client.answer(msg,"ko","No Such Device")
                return
            dev = self.devices[data["device"]]
            board = self._get_board(dev)
            if not board:
                self.client.answer(msg,"ko","No Such Board")
                return
        if "clean" in data:
            #clean compiler cache
            compiler.cncache.clear()
            self.client.answer(msg,"ok")
            return

        try:
            bcode,cobjs=self._compile_proj(proj,board)
            self.bcodes[proj["id"]]=bcode

        except compiler.CError as ce:
            errs = [{"line":ce.line,"msg":ce.toCleanStr()}]
            self.client.answer(msg,{
                "result":{
                    self.stripfname(proj,ce.filename): {
                        "E":errs,
                        "W":[]
                        }
                    }
                })
            ide.viper.say("COMPILER","in "+self.stripfname(proj,ce.filename)+" at line "+str(errs[0]["line"])+": "+ce.toCleanStr())
            if isinstance(ce,compiler.CModuleNotFound):
                mnames = ce.module.split(".")
                while mnames:
                    mname = ".".join(mnames)
                    if mname in tb.i_modules_info:
                        ide.viper.say("COMPILER","library "+mname+" seems installed correctly: check if it is also enabled")
                        break
                    elif mname in tb.modules_info:
                        ide.viper.say("COMPILER","library "+mname+" exists but it is not installed: search for it in the viper package manager")
                        break
                    mnames.pop()
                else:
                    ide.viper.say("COMPILER","module "+ce.module+" unknown")
            return
        except Exception as e:
            logger.error("COMPILER: %s",ide.viper.pretty_exc(e))
            self.client.answer(msg,"ko","General Exception! "+str(e))
            return
        js = []
        dirs = [self.dname(proj),compiler.libdir]
        for cb in cobjs:
            cs = cb.toDict(dirs)
            js.append(cs)
        self.client.answer(msg,{"info":js})

    def upload(self,msg):
        data = msg["data"]
        proj = self.projects[data["id"]]
        if data["device"] not in self.devices:
            self.client.answer(msg,"ko","No Such Device")
        dev = self.devices[data["device"]]
        board = self._get_board(dev)
        if not board:
            self.client.answer(msg,"ko","No Such Board")
            return
        if data["device"] in self.consoles:
            self.say("Stopping Console...")
            self.consoles[data["device"]].stop()
            self.consoles[data["device"]].has_close.wait()
            del self.consoles[data["device"]]
            self.say_close_console(data["device"])

        #if not self.bcodes.get(proj["id"],None):
        try:
            bcode,cobjs = self._compile_proj(proj,board)
            self.bcodes[proj["id"]]=bcode
        except compiler.CError:
            self.client.answer(msg,"ko","Your program has errors! Solve them before uploading...")
            return
        except Exception as e:
            logger.error("UPLOADER: %s",ide.viper.pretty_exc(e))
            self.client.answer(msg,"ko","Something wrong while compiling: "+ide.viper.pretty_exc(e))
            return
        try:
            upl = uplinker.Uplinker(dev["port"],self.bcodes[proj["id"]],self.say)
            resetmode = board.manual_reset
            if resetmode==0:
                #no user interaction
                #Arduino Due
                board.reset()
                uplret = upl.upload(addr=board.addr)
            elif resetmode==1:
                #board must be reset before upload opens port
                #Spark
                self.say("\n"+("#"*80)+"\n#       Please RESET your "+dev["name"]+" in the next 5 seconds\n"+("#"*80)+"\n")
                self.notification("Please RESET your "+dev["name"]+" in the next 5 seconds")
                time.sleep(5)
                uplret = upl.upload(addr=board.addr)
            elif resetmode==2:
                #board must be reset after upload opens port
                #ST Nucleo
                self.say("\n"+("#"*80)+"\n#       Please RESET your "+dev["name"]+" in the next 5 seconds\n"+("#"*80)+"\n")
                self.notification("Please RESET your "+dev["name"]+" in the next 5 seconds")
                time.sleep(2)
                uplret = upl.upload(addr=board.addr)
        except uplinker.UplinkException as ue:
            logger.error("UPLOADER: %s",ide.viper.pretty_exc(ue))
            self.say("[UPLOAD]: ooops, uploading failed with the following error: %s"%ue.msg)
            self.client.answer(msg,"ko",ue.msg)
            ide.viper.getStatListener().send_stat("upload",{"error":ue.msg,"remoteId":proj.get("remoteId","")})
            return
        except Exception as e:
            logger.error("UPLOADER: %s",ide.viper.pretty_exc(e))
            self.say("[UPLOAD]: ooops, uploading failed with the following error: %s"%ide.viper.pretty_exc(e))
            self.client.answer(msg,"ko",ide.viper.pretty_exc(e))
            ide.viper.getStatListener().send_stat("upload",{"exception":ide.viper.pretty_exc(e),"remoteId":proj.get("remoteId","")})
            return
        ide.viper.getStatListener().send_stat("upload",{"device":dev["device"],"board":dev["board"],"vm":uplret,"remoteId":proj.get("remoteId","")})
        js = []
        dirs = [self.dname(proj),compiler.libdir]
        for cb in cobjs:
            cs = cb.toDict(dirs)
            js.append(cs)
        self.client.answer(msg,{"info":js})

    # def synch_project(self,msg):
    #     data = msg["data"]
    #     proj = self.projects[data["id"]]
    #     localrepo = self.dname(proj)
    #     reponame = data["repo"]
    #     if not isinstance(reponame,str):
    #         repo = self._get_repo(proj,reponame)
    #         origin = self.git.origin(repo["url"],repo["user"],repo["pass"])
    #     else:
    #         origin = ide.viper.viper_origin#self.git.origin(data["repo"]["url"],data["repo"]["user"],data["repo"]["pass"])
    #         repo = "_"
    #     ide.viper.say("SYNC","Commiting to local")
    #     try:
    #         self.git.do_commit(localrepo,"Automatic_Commit")
    #     except ide.vipergit.GitException as ge:
    #         logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
    #         self.client.answer(msg,"ko","Can't commit project locally! "+str(ge))
    #         ide.viper.say("SYNC","oops, can't commit! "+str(ge))
    #         return
    #     if self.backend.logged:
    #         try:
    #             ide.viper.say("SYNC","Checking remote status")
    #             if self.git.remote_update(origin,localrepo):
    #                 ide.viper.say("SYNC","Remote is ahead of local...let's merge")
    #                 self.git.merge(origin,localrepo)
    #             ide.viper.say("SYNC","Pushing to remote")
    #             eee,out = self.git.push(localrepo,origin)
    #             ide.viper.say("SYNC",out)
    #         except ide.vipergit.GitException as ge:
    #             logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
    #             self.client.answer(msg,"ko","Can't push project remotely! "+str(ge))
    #             ide.viper.say("SYNC","oops, can't push to cloud! "+str(ge))
    #             self.fill_from_path(proj)
    #             self.send_project(proj)
    #             return
    #     self.fill_from_path(proj)
    #     self.send_project(proj)
    #     ide.viper.say("SYNC","Everything done!")
    #     self.client.answer(msg,"ok")

    def push_project(self,msg):
        if not self.backend.logged:
            self.client.answer(msg,"ko","You are not logged!")
            return

        data = msg["data"]
        proj = self.projects[data["id"]]
        localrepo = self.dname(proj)
        reponame = data["repo"]
        if not isinstance(reponame,str):
            repo = self._get_repo(proj,reponame)
            origin = self.git.origin(repo["url"],repo["user"],repo["pass"])
        else:
            origin = ide.viper.viper_origin#self.git.origin(data["repo"]["url"],data["repo"]["user"],data["repo"]["pass"])
            repo = "_"
        ide.viper.say("SYNC","Commiting to local")
        try:
            self.git.do_commit(localrepo,"Automatic_Commit")
        except ide.vipergit.GitException as ge:
            logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
            self.client.answer(msg,"ko","Can't commit project locally! "+str(ge))
            ide.viper.say("SYNC","oops, can't commit! "+str(ge))
            return
        if self.backend.logged:
            ee = self._new_proj_cloud_ready(proj)
            if ee>=0:
                try:
                    self.git.add_remote(origin,self.backend.get_project_repo_url(proj),localrepo)
                    ide.viper.say("SYNC","Pushing to remote")
                    self.git.push_whatever(localrepo,origin)
                except ide.vipergit.GitException as ge:
                    logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
                    self.client.answer(msg,"ko","Can't push project remotely! "+str(ge))
                    ide.viper.say("SYNC","oops, can't push to cloud! "+str(ge))
                    self.fill_from_path(proj)
                    self.send_project(proj)
                    return
            elif ee<=-2:
                logger.error("Something wrong with local commit")
                self.client.answer(msg,"ko","Something wrong with local commit")
                return
            else:
                logger.error("Can't create project remotely")
                self.client.answer(msg,"ko","Can't create project remotely!")
                return
        self.fill_from_path(proj)
        self.send_project(proj)
        self.store()

        ide.viper.say("SYNC","Everything done!")
        self.client.answer(msg,"ok")

    def pull_project(self,msg):
        if not self.backend.logged:
            self.client.answer(msg,"ko","You are not logged!")
            return
        data = msg["data"]
        proj = self.projects[data["id"]]
        localrepo = self.dname(proj)
        reponame = data["repo"]
        if not isinstance(reponame,str):
            repo = self._get_repo(proj,reponame)
            origin = self.git.origin(repo["url"],repo["user"],repo["pass"])
        else:
            origin = ide.viper.viper_origin#self.git.origin(data["repo"]["url"],data["repo"]["user"],data["repo"]["pass"])
            repo = "_"
        ide.viper.say("SYNC","Commiting to local")
        try:
            self.git.do_commit(localrepo,"Automatic_Commit")
        except ide.vipergit.GitException as ge:
            logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
            self.client.answer(msg,"ko","Can't commit project locally! "+str(ge))
            ide.viper.say("SYNC","oops, can't commit! "+str(ge))
            return
        if self.backend.logged:
            ee = self._new_proj_cloud_ready(proj)
            if ee>=0:
                try:
                    #update remote
                    self.git.add_remote(origin,self.backend.get_project_repo_url(proj),localrepo)
                    ide.viper.say("SYNC","Checking remote status")
                    #it may not have a remote
                    self.git.remote_update(origin,localrepo)
                    self.git.reset_to_master(localrepo,origin)
                except ide.vipergit.GitException as ge:
                    logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
                    self.client.answer(msg,"ko","Can't push project remotely! "+str(ge))
                    ide.viper.say("SYNC","oops, can't push to cloud! "+str(ge))
                    self.fill_from_path(proj)
                    self.send_project(proj)
                    self.store()
                    return
            elif ee<=-2:
                logger.error("Something wrong with local commit")
                self.client.answer(msg,"ko","Something wrong with local commit")
                return
            else:
                logger.error("Can't create project remotely")
                self.client.answer(msg,"ko","Can't create project remotely!")
                return

        self.fill_from_path(proj)
        self.send_project(proj)
        self.store()
        ide.viper.say("SYNC","Everything done!")
        self.client.answer(msg,"ok")


    def synch_missing(self,msg):
        if self.backend.logged:
            rprojs = self.backend.get_projects()
            mprojs = {}
            not_downloaded=[]
            not_uploaded=[]
            downloaded=0
            uploaded=0
            not_updated=[]
            diverged = []
            updated=0
            for name,proj in self.projects.items():
                if "remoteId" in proj:
                    mprojs[proj["remoteId"]]=proj
            #remote synch
            for rproj in rprojs:
                if rproj["remoteId"] not in mprojs:
                    # a remote project not present locally: download
                    ide.viper.say("SYNC","Downloading remote project "+rproj["name"])
                    proj = {
                        "name":rproj["name"],
                        "remoteId":rproj["remoteId"],
                        "desc":rproj["desc"],
                        "tags":rproj["tags"],
                        "repos":rproj["repos"],
                        "files":rproj["files"]
                    }
                    uid = self._new_project_dir(proj)
                    proj["synched"]="uptodate"
                    try:
                        self.git.clone(self.dname(proj),self.backend.get_project_repo_url(proj),ide.viper.viper_origin)
                    except ide.vipergit.GitException as ge:
                        logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
                        proj["synched"]="unknown"
                        shutil.rmtree(self.dname(proj),ignore_errors=True)
                        not_downloaded.append(proj["name"])
                        ide.viper.say("SYNC","Not downloaded...")
                        continue
                    self.fill_from_path(proj)
                    self.fill_project(proj)
                    self._save_project(proj)
                    self.projects[uid]=proj
                    ide.viper.say("SYNC","Ok!")
                    downloaded+=1
                else:
                    #a local project present also remotely: update by merging current version
                    proj = mprojs[rproj["remoteId"]]
                    ide.viper.say("SYNC","Checking local project "+rproj["name"])
                    try:
                        self.git.do_commit(self.dname(proj),"Sync_Commit")
                        self.git.add_remote(ide.viper.viper_origin,self.backend.get_project_repo_url(proj),self.dname(proj))
                        status = self.git.check_status(self.dname(proj),ide.viper.viper_origin)
                        if status == "behind":
                            ide.viper.say("SYNC","The cloud version of "+rproj["name"]+" is newer, downloading!")
                            self.git.reset_to_master(self.dname(proj),ide.viper.viper_origin)
                            updated+=1
                            proj["synched"]="uptodate"
                        elif status == "ahead":
                            ide.viper.say("SYNC","The cloud version of "+rproj["name"]+" is older, uploading!")
                            self.git.push(self.dname(proj),ide.viper.viper_origin)
                            updated+=1
                            proj["synched"]="uptodate"
                        elif status == "uptodate":
                            ide.viper.say("SYNC","The cloud version of "+rproj["name"]+" is the same as local, nothing to do")
                            proj["synched"]="uptodate"
                        elif status == "diverged":
                            ide.viper.say("SYNC","Ooops, local and cloud versions of "+rproj["name"]+" have diverged!? You must check manually!")
                            diverged.append(proj["name"])
                            proj["synched"]="diverged"
                    except ide.vipergit.GitException as ge:
                        proj.synched="unknown"
                        logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
                        not_updated.append(proj["name"])
            #local synch
            for name,proj in self.projects.items():
                if "remoteId" in proj:
                    continue
                ide.viper.say("SYNC","Uploading local project "+name)
                try:
                    ee=self._new_proj_cloud_ready(proj)
                    if ee>=0:
                        self.git.add_remote(ide.viper.viper_origin,self.backend.get_project_repo_url(proj),self.dname(proj))
                        self.git.push(self.dname(proj),ide.viper.viper_origin)
                        ide.viper.say("SYNC","Ok!")
                        uploaded+=1
                    elif ee<0:
                        not_uploaded.append(proj["name"])
                        ide.viper.say("SYNC","Not uploaded")
                except ide.vipergit.GitException as ge:
                    not_uploaded.append(proj["name"])
                    logger.error("Something wrong with git: %s",ide.viper.pretty_exc(ge))
                    ide.viper.say("SYNC","Not uploaded")

            self.store()
            if not self.send_last_project():
                self.send_projects()
            ide.viper.say("SYNC","New projects downloaded: "+str(downloaded))
            ide.viper.say("SYNC","Local projects sync'd: "+str(uploaded))
            if not_downloaded:
                ide.viper.say("SYNC","Failed to download: "+str(not_downloaded))
            if not_uploaded:
                ide.viper.say("SYNC","Failed to sync: "+str(not_uploaded))
            if diverged:
                ide.viper.say("SYNC","You must manually check: "+str(diverged))
            ide.viper.say("SYNC","Done!")
            if (diverged):
                self.client.answer(msg,"ok",str(diverged))
            else:
                self.client.answer(msg,"ok","")
        else:
            self.client.answer(msg,"ko","You are not logged!")

    def _new_proj_cloud_ready(self,proj):
        if "remoteId" not in proj:
            #a local project not present remotely: upload
            res = self.backend.new_project(proj)
            self.store()
            if res:
                try:
                    proj["remoteId"]=res["remoteId"]
                    proj["synched"]="uptodate"
                    self.git.add_remote(ide.viper.viper_origin,self.backend.get_project_repo_url(proj),self.dname(proj))
                    self.git.do_commit(self.dname(proj),"Sync_Commit")
                    return 1
                except ide.viper.GitException as ge:
                    del proj["remoteId"]
                    proj["synched"]="unknown"
                    return -2
            else:
                return -1
        return 0

    def _parse_linter(self,out):
        rline = "([A-Z]):(.*):([0-9]+):([0-9]+):([0-9A-Z]+): (.*)"
        matcher = re.compile(rline)
        lines = out.split("\n")
        res = {}
        prev_matched = False
        prev_ended = True
        pres={}
        for line in lines:
            if line.startswith("Your code has been rated"):
                if "(" in line:
                    line = line[:line.find("(")]
                res["Rate"]=line
                continue
            mth = matcher.match(line)
            #print("line:",line,mth)
            if mth:
                prev_matched = True
                prev_ended = line.endswith("_-_-_")
                fname = mth.group(2)
                if fname not in res:
                    res[fname]={"W":[],"C":[],"E":[],"F":[],"R":[]}
                pres = {
                    "line":int(mth.group(3)),
                    "col":int(mth.group(4)),
                    "type":mth.group(5),
                    "msg":mth.group(6).replace("_-_-_","")
                    }
                res[fname][mth.group(1)].append(pres)
            else:
                if prev_matched and pres and not prev_ended:
                    if line.endswith("_-_-_"):
                        #line is made of spaces and ^ only
                        line.replace("_-_-_","")
                        cc = line.count("^")
                        cl=[0]
                        for i in range(cc):
                            cl.append(line.index("^",cl[-1]+1))
                        cl.pop(0)
                        pres["dots"]=cl
                        #print(pres)
                        prev_matched=False
                else:
                    prev_matched = False
        return res

    def lint_project(self,msg):
        data=msg["data"]
        if data["id"] in self.projects:
            proj = self.projects[data["id"]]
            #res,output = vtools.run("linter","-f", "parseable", "--persistent=n",("--msg-template='{C}:{path}:{line}:{column}:{msg_id}: {msg}'",),"main.py",cwd=self.dname(proj))
            res,output = tb.run("linter","--rcfile="+os.path.join(vconf.tooldir,"pylintrc"),"main.py",cwd=self.dname(proj))
            #print(output)
            #print("--rcfile="+os.path.join(vtools.conf.instdir,"pylintrc"))
            self.client.answer(msg,"ok",self._parse_linter(output))
        else:
            self.client.answer(msg,"ko","Project does not exists!")

    def vpm_search(self,msg):
        data = msg["data"]
        try:
            self.vpm.set_logger(lambda x: ide.viper.say("VPM",x))
            #todo: add search options
            result = self.vpm.search(data["query"])
            tresult = self.vpm.search(data["query"],testing=True)
            result.extend(tresult)
            self.vpm.set_logger(None)
            self.client.answer(msg,{"result":[x.to_versioned_dict(x.get_last_version()) for x in result]})
        except Exception as e:
            self.client.answer(msg,"ko","Something wrong: "+str(e))

    def vpm_cmd(self,msg):
        data = msg["data"]
        self.vpm.set_credentials_fn(self.backend.get_credentials)
        self.vpm.set_logger(lambda x: ide.viper.say("VPM",x))

        try:
            if data["cmd"]=="install":
                if not self.backend.logged:
                    self.client.answer(msg,"ko","You must be logged to install packages")
                else:
                    vrs = self.vpm.to_version(data["version"])
                    res = self.vpm.install(self.vpm.get_package(by=data["uid"],version=vrs),vrs)
                    if res is None:
                        self.vpm.persist()
                        self.client.answer(msg,"ok","")
                    else:
                        self.client.answer(msg,{"conflicts":res},"There are conflicts!")
                    self.backend.dload_package(data["uid"])
                    self.backend.sync_pkglist()
            elif data["cmd"]=="uninstall":
                if not self.backend.logged:
                    self.client.answer(msg,"ko","You must be logged to uninstall packages")
                else:
                    vrs = self.vpm.to_version(data["version"])
                    ppk = self.vpm.get_package(by=data["uid"],version=vrs)
                    res = self.vpm.uninstall(ppk,force=True)
                    self.vpm.persist()
                    self.client.answer(msg,"ok","")
                    self.backend.sync_pkglist()
            elif data["cmd"]=="install_info":
                vrs = self.vpm.to_version(data["version"])
                res = self.vpm.install_info(self.vpm.get_package(by=data["uid"],version=vrs),vrs)
                if res:
                    self.client.answer(msg,{"result":res},"ok")
                else:
                    self.client.answer(msg,"ko","No such package")
            elif data["cmd"]=="uninstall_info":
                vrs = self.vpm.to_version(data["version"])
                pkg = self.vpm.get_package(by=data["uid"],version=vrs,installed=True)
                res = self.vpm.uninstall_info(pkg,ext=True)
                if res:
                    self.client.answer(msg,{"result":{"package":pkg.to_detailed_versioned_dict(pkg.installed),"rem":res["rem"],"has_core":res["has_core"]}},"ok")
                else:
                    self.client.answer(msg,"ko","No such package")
            elif data["cmd"]=="sync":
                self.vpm.update_from_repos()
                self.client.answer(msg,"ok","")
            elif data["cmd"]=="enable":
                vrs = self.vpm.to_version(data["version"])
                ppk = self.vpm.get_package(by=data["uid"],version=vrs)
                self.vpm.enable_package(ppk)
                self.vpm.persist()
                self.client.answer(msg,"ok","")
            elif data["cmd"]=="disable":
                vrs = self.vpm.to_version(data["version"])
                ppk = self.vpm.get_package(by=data["uid"],version=vrs)
                self.vpm.disable_package(ppk)
                self.vpm.persist()
                self.client.answer(msg,"ok","")
            elif data["cmd"]=="rate":
                if self.backend.rate_package(data["uid"],data["version"]):
                    self.client.answer(msg,"ok","")
                else:
                    self.client.answer(msg,"ko","Backend Error")
            elif data["cmd"]=="info":
                vrs = self.vpm.to_version(data["version"])
                res = self.vpm.get_package(by=data["uid"],version=vrs)
                if res:
                    self.client.answer(msg,{"package":res.to_detailed_versioned_dict(vrs)},"ok")
                else:
                    self.client.answer(msg,"ko","no such package")
            elif data["cmd"]=="update_all":
                self.vpm.update_all()
                self.client.answer(msg,"ok","Everything is up to date")
        except Exception as e:
            self.client.answer(msg,"ko","Something wrong: "+str(e))
            logger.error("VPM_CMD %s %s",str(msg),ide.viper.pretty_exc(e))
        self.vpm.set_credentials_fn(None)
        self.vpm.set_logger(None)
        self.send_vpm()

    def namespace_cmd(self,msg):
        data = msg["data"]
        # self.vpm.set_credentials_fn(self.backend.get_credentials)
        # self.vpm.set_logger(lambda x: ide.viper.say("VPM",x))

        try:
            if data["cmd"]=="list":
                if not self.backend.logged:
                    self.client.answer(msg,"ko","You must be logged to query namespaces")
                else:
                    res = self.backend.list_namespaces()
                    self.client.answer(msg,res,"ok")
            elif data["cmd"]=="create":
                if not self.backend.logged:
                    self.client.answer(msg,"ko","You must be logged to create a namespace")
                else:
                    res = self.backend.create_namespace(data["args"])
                    if res is None:
                        self.client.answer(msg,"ko","Not possible to create such a namespace")
                    else:
                        self.client.answer(msg,res,"ok")
        except Exception as e:
            self.client.answer(msg,"ko","Something wrong: "+str(e))
            logger.error("NSP_CMD %s %s",str(msg),ide.viper.pretty_exc(e))

    def send_vpm_check_updated(self):
        try:
            if self.backend.connected:
                self.rrepo,self.trepo = self.vpm.are_repo_updated()
                if not self.rrepo:
                    ide.viper.say("VPM","Official repository has been updated! Sync to get it...")
                if not self.trepo:
                    ide.viper.say("VPM","Testing repository has been updated! Sync to get it...")
                self.send_status()
        except Exception as e:
            raise e
            logger.error("Exception while checking for updated repos %s"%str(e))

    def refresh(self,msg):
        tb.load_modules_info()
        self.client.answer(msg,"ok")
        self.send_examples()
        self.hwlistener.refresh()
        self.send_devices()
        self.send_vms()
        self.send_status()
        self.send_projects()


    def restart(self,msg):
        self.client.answer(msg,"ok")
        ide.viper.restart=tb.tools["restart"]
        ide.viper.getLooper().running=False

    def dispatch(self,msg):
        msgtype = msg["type"]
        #logger.info("dispatching message %s",str(msg))

        if hasattr(self,msgtype):
            try:
                getattr(self,msgtype)(msg)
            except Exception as e:
                self.client.answer(msg,"ko","Fatal Error "+ide.viper.pretty_exc(e))
                logger.error("DISPATCH %s %s",str(msg),ide.viper.pretty_exc(e))
        else:
            logger.error("Unknown message %s",msgtype)

    def makedoc(self,msg):
        data = msg["data"]
        ide.viper.say("DOCS","Preparing to build documentation...")
        path = os.path.join(self.projdir,data['id'])
        # basedir=os.path.split(os.path.realpath(__file__))[0]
        basedir = os.path.join(vconf.envdirs["core"],"viper")
        pth = os.path.realpath(path)
        docpath = os.path.join(pth,"docs")
        htmlpath = os.path.join(docpath,"html")
        shutil.copy(os.path.join(basedir,"docconf.py"),os.path.join(docpath,"conf.py"))
        shutil.rmtree(htmlpath,ignore_errors=True)
        ide.viper.say("DOCS","Building...")
        res,out=tb.run("sphinx","-b","html",docpath,htmlpath)
        ide.viper.say("DOCS",out)
        if res:
            ide.viper.say("DOCS","Error!")
        else:
            ide.viper.say("DOCS","Done")
        os.remove(os.path.join(docpath,"conf.py"))
        os.remove(os.path.join(docpath,"layout.html"))
        shutil.copy(os.path.join(basedir,"viper.css"),os.path.join(docpath,"html","_static","viper.css"))
        self.client.answer(msg,"ok")


    def toggle_lib_rt(self,msg):
        data = msg["data"]
        if data["id"] in self.projects:
            proj = { data["id"]: self.projects[data["id"]] }
            if proj[data["id"]]["mode"] == "testing":
                proj[data["id"]]["mode"] = "release"
            else:
                proj[data["id"]]["mode"] = "testing"
            self.db.store_projects(proj)
            self.client.answer(msg,"ok")
            return
        self.client.answer(msg,"ko","Project does not exists!")

    def publish_library(self,msg):
        data = msg["data"]
        if data["id"] in self.projects:
            proj = self.projects[data["id"]]

            with open(self.fname(proj,'package.json'),"w") as ff:
                pkg_json = { 'name' : proj["name"], 'dependencies': {}, 'examples': [] }
                for dep, version in data["dependencies"].items():
                    pkg_json['dependencies'].update({ dep: version })
                for example_p in data["examples"]:
                    if example_p in self.projects:
                        pkg_json['examples'].append(self.projects[example_p]["name"].replace(" ", "_"))
                ff.write(json.dumps(pkg_json))

            gb = github.GitHub()
            liburl = "github://"+data["gh_user"]+"/"+data["gh_repo"]

            package_name = "community." + proj["namespace"] + "." + proj["mod_name"]
            my_pkgs = self.backend.list_packages()
            if not any([ pkg['fullname'] == package_name for pkg in my_pkgs ]):
                res = self.backend.create_package(proj["name"],proj["desc"],liburl,proj["namespace"],proj["mod_name"],"community",proj.get("keywords",""))

                if not res:
                    self.client.answer(msg,"ko","Can't create package!")
                    return

                pkgkey = res
                gb.auth(data["gh_user"],data["gh_psw"])
                try:
                    gb.create_repo(data["gh_repo"])
                    gb.create_hook(data["gh_repo"],vconf.viper_server+"githubhook","create",hook_secret=pkgkey)
                except viper.github.GitHubException:
                    pass

            tmpdir = os.path.join(vconf.envdirs["tmp"],"_tmp"+data["id"])
            shutil.rmtree(tmpdir,onerror=self.vpm.remove_readonly)

            res,out = vutils.run("git","clone",liburl.replace("github://","https://github.com/"),tmpdir,cwd=vconf.envdirs["tmp"])
            if res:
                self.client.answer(msg,"ko","Can't download package: "+out)
                return

            vrs_err = False
            version = proj["mode"][0] + data["version"]
            res,out = vutils.run("git","tag","--list",cwd=tmpdir)
            v_tags  = [ v for v in out.split("\n") if v != '' ]
            try:
                rcv_version   = self.vpm.to_version(version)
                repo_versions = [ self.vpm.to_version(tag) for tag in v_tags if tag[0] == proj["mode"][0] ]
            except Exception:
                self.client.answer(msg,"ko","wrong package version")
                vrs_err = True
            if (len(repo_versions)!= 0):
                repo_versions.sort()
                if repo_versions[-1] >= rcv_version:
                    self.client.answer(msg,"ko","new version must be higher")
                    vrs_err = True

            if vrs_err:
                shutil.rmtree(tmpdir,onerror=self.vpm.remove_readonly)
                return

            for to_del in os.listdir(tmpdir):
                if to_del != '.git' and to_del != '.gitignore':
                    try:
                        os.remove(os.path.join(tmpdir, to_del))
                    except IsADirectoryError:
                        shutil.rmtree(os.path.join(tmpdir, to_del),onerror=self.vpm.remove_readonly)

            for to_cp in os.listdir(self.dname(proj)):
                if to_cp != '.git' and to_cp != '.gitignore' and to_cp != 'main.py':
                    try:
                        shutil.copy(self.fname(proj, to_cp),tmpdir)
                    except IsADirectoryError:
                        shutil.copytree(self.fname(proj, to_cp),os.path.join(tmpdir, to_cp))

            examples_dir = os.path.join(tmpdir,'examples')

            ex_ff = open(os.path.join(examples_dir,'order.txt'),"w")
            ex_ff.write("#" + proj["namespace"] + "\n")
            ex_ff.write("\t##" + proj["name"].replace(" ", "_") + "\n")

            for example_p in data["examples"]:
                if example_p in self.projects:
                    ex_ff.write("\t\t" + self.projects[example_p]["name"].replace(" ", "_") + "\n")
                    example_dir = os.path.join(examples_dir,self.projects[example_p]["name"].replace(" ", "_"))
                    os.mkdir(example_dir)
                    for to_cp in os.listdir(self.dname(self.projects[example_p])):
                        if to_cp != '.git' and to_cp != '.gitignore':
                            try:
                                shutil.copy(self.fname(self.projects[example_p], to_cp),example_dir)
                            except IsADirectoryError:
                                shutil.copytree(self.fname(self.projects[example_p], to_cp),os.path.join(example_dir, to_cp))

            ex_ff.close()

            with open(os.path.join(tmpdir,".gitignore"),"w") as gg:
                gg.write("*.DS_Store\n")
                gg.write("*.doctrees\n")

            res,out = vutils.run("git","add","-A",'.',cwd=tmpdir)

            # TODO: check res
            res,out = vutils.run("git","commit","-m",data["changelog"],cwd=tmpdir)

            res,out = vutils.run("git","tag",version,cwd=tmpdir)

            new_origin = liburl.replace("github://","https://"+data["gh_user"]+":"+data["gh_psw"]+"@github.com/")
            res,out = vutils.run("git","config","remote.origin.url",new_origin,cwd=tmpdir)

            res,out = vutils.run("git","push","origin","master",version,cwd=tmpdir)

            shutil.rmtree(tmpdir,onerror=self.vpm.remove_readonly)
            self.client.answer(msg,"ok")
            return
        self.client.answer(msg,"ko","Project does not exists!")

    def get_dependencies(self,msg):
        data = msg["data"]
        if data["id"] in self.projects:
            proj = self.projects[data["id"]]
            res  = { 'ok': {}, 'fail': [] }

            for py_file in proj["files"]:
                if py_file.endswith(".py") and py_file!="main.py":
                    # res.update(self._find_imported_packages(proj, py_file))
                    r = self._find_imported_packages(proj, py_file)
                    print("checking",py_file)
                    print("--->",r)
                    res['ok'].update(r['ok'])
                    res['fail'] += ( f for f in r['fail'] if not f in res['fail'])

            self.client.answer(msg,res,"ok")
            return
        self.client.answer(msg,"ko","Project does not exists!")

    def get_my_pkg_version(self,msg):
        data = msg["data"]

        if not data["id"] in self.projects:
            self.client.answer(msg,"ko","Project does not exists!")
            return

        proj = self.projects[data["id"]]
        package_name = "community." + proj["namespace"] + "." + proj["mod_name"]
        my_pkgs = self.backend.list_packages()

        if any([ pkg['fullname'] == package_name for pkg in my_pkgs ]):
            my_pkg  = [ pkg for pkg in my_pkgs if pkg['fullname'] == package_name ][0]
            gb = github.GitHub()
            tgs = gb.get_taglist(my_pkg['url'].split('/')[-1], my_pkg['url'].split('/')[-2])
            if not tgs:
                # should not be possible
                self.client.answer(msg,"ko","package not published")
                return
            else:
                try:
                    vp_versions = [ self.vpm.to_version(tag) for tag in tgs if tag[0] == proj["mode"][0] ]
                except Exception:
                    self.client.answer(msg,"ko","wrong package version")
                    return
                vp_versions.sort()
                res = {'version': str(vp_versions[-1]), 'fullname': package_name}
                self.client.answer(msg,res,"ok")
                return
        self.client.answer(msg,"ko","package not published")

    def change_lib_namespace(self,msg):
        data = msg["data"]
        if data["id"] in self.projects:
            proj = { data["id"]: self.projects[data["id"]] }
            if proj[data["id"]]["local"] == True:
                proj[data["id"]]["local"] = False
                proj[data["id"]]["namespace"] = data["namespace"]
            self.db.store_projects(proj)
            self.backend.update_project(proj)
            self.client.answer(msg,"ok")
            self.send_project(proj)
            return
        self.client.answer(msg,"ko","Project does not exists!")


####### SERVER TO CLIENT

    def say(self,msg):
        self.client.send({"type":"msg","data":msg})

    def say_console(self,console,msg):
        self.client.send({"type":"console_msg","data":{"console_id":console,"msg":self.data_encode(msg)}})

    def say_close_console(self,console):
        self.client.send({"type":"console_close","data":{"console_id":console}})

    def dialog(self,title,msg,ok=True):
        self.client.send({"type":"dialog","data":{"title":title,"msg":msg,"ok":ok}})

    def notification(self,text,style="viperwarning",position="bottom center"):
        self.client.send({"type":"notify","data":{"msg":text,"position":position,"style":style}})

    def send_status(self):
        uname,upass,email = self.db.get_user()
        docvrs = 0
        docvpm = 0
        vrs = self.vpm.version(vconf.testing)
        try:
            dd = self.vpm.get_package(by="core.viper.docs",version=vrs)
            docvrs = str(dd.installed)
        except:
            pass
        try:
            dd = self.vpm.get_package(by="core.viper.vpm",version=vrs)
            docvpm = str(dd.installed)
        except:
            pass

        self.client.send({"type":"status","data":{
            "version":vconf.get_viper_version(),
            "build":vconf.get_viper_build(),
            "logged":self.backend.logged,
            "user":uname,
            "email":email,
            "password":upass,
            "update":ide.viper.getUpdater().has_update(),
            "update_version":"...",
            "full":vconf.get_viper_version_string(),
            "msg":"...",
            "repostatus":[self.rrepo,self.trepo],
            "connected":self.backend.connected,
            "docvrs": docvrs,
            "docvpm": docvpm
            }})

    def send_last_project(self):
        cprj = self.db.get_current_project()
        if cprj and cprj in self.projects:
            self.fill_from_path(self.projects[cprj])
            self.send_projects()
            self.send_project(self.projects[cprj])
            return True
        return False

    def send_project(self,proj,msg=True):
        self.fill_project(proj)
        proj["last_access"] = datetime.datetime.utcnow().isoformat()
        logger.info("sending project %s",proj["name"])
        prj = copy.deepcopy(proj)
        self.client.send({"type":"project","data":prj})
        self.unfill_project(proj)
        if msg:
            ide.viper.say("PROJECT","Here is your project "+proj["name"])
        self.db.set_current_project(proj)

    def send_projects(self):
        data = []
        for uid,proj in self.projects.items():
            prj = copy.deepcopy(proj)
            data.append(prj)
        self.client.send({"type":"projects","data":data})

    def _get_ex_desc(self,root,exname):
        desc  = ""
        if os.path.exists(os.path.join(root,exname,"project.md")):
            with open(os.path.join(root,exname,"project.md"),"r") as exf:
                desc = exf.read()
            lines = desc.split("\n")
            for line in reversed(lines):
                if line.startswith("groups:"):
                    desc=desc.replace(line,"")
                if line.startswith("tags:"):
                    desc=desc.replace(line,"")
                    break
        return desc

    def _get_ex_code(self,root,exname):
        if os.path.exists(os.path.join(root,exname,"main.py")):
            with open(os.path.join(root,exname,"main.py"),"r") as exf:
                return exf.read()
        return ""


    def send_examples(self):
        stack = []
        extree = {}
        exlist = []
        exord = []
        exn=0
        grpn=0
        frst=0
        for exdir in tb.get_example_dirs():
            try:
                #print("trying",os.path.join(exdir,"order.txt"))
                with open(os.path.join(exdir,"order.txt")) as ff:
                    lines = ff.readlines()
                for line in lines:
                    line = line.strip()
                    if not line or len(line)<4 or line.startswith(";"):
                        continue
                    pos = line.count("#")-1
                    if pos>=0:
                        if frst and (not stack or stack[0]!="Libraries"):
                            stack=["Libraries"]
                            extree[("|".join(stack))]=[]
                            exord.append(("|".join(stack)))
                        stack[pos+frst:]=[line[(pos+1):]]
                        extree[("|".join(stack))]=[]
                        exord.append(("|".join(stack)))
                        continue
                    else:
                        ex = {}
                        ex["desc"]= self._get_ex_desc(exdir,line)
                        ex["code"]= self._get_ex_code(exdir,line)
                        ex["name"] = line.replace("_"," ")
                        ex["repo"] = os.path.join(exdir,line)
                        ex["tags"]= []
                        ex["last_access"]="{0:#0{1}x}".format(exn,6)
                        ex["id"] = str(uuid.uuid4())
                        ex["num"]=exn
                        exlist.append(ex)
                        extree[("|".join(stack))].append(exn)
                        exn+=1
                frst=1
            except Exception as e:
                print(e)
                pass
        self.client.send({"type":"examples","data":[extree,exlist,exord]})

    def send_devices(self):
        ddata = []
        if self.devices:
            for k,x in self.devices.items():
                y = copy.deepcopy(x)
                if k in self.dev_props:
                    y["name"]=self.dev_props[k]["name"]
                    y["desc"]=self.dev_props[k]["desc"]
                #y["virtual"]=False
                ddata.append(y)
        self.client.send({"type":"devices","data":ddata})

    def send_vpm(self):
        res = [x.to_versioned_dict(x.installed) for x in self.vpm.get_all_installed_packages()]
        mdn = [x.to_versioned_dict(v) for x,v in self.vpm.query_packages(("","",1,"order by downloads desc limit 0,20")) if not x.installed]
        mrt = [x.to_versioned_dict(v) for x,v in self.vpm.query_packages(("type","==","lib","order by rating desc, rating_count desc limit 0,40")) if not x.installed]
        mcr = [x.to_versioned_dict(v) for x,v in self.vpm.query_packages(("last_update",">",0,"order by last_update desc limit 0,40")) if not x.installed]
        tb.load_modules_info()
        iml = list(tb.i_modules_info.keys())
        iml.sort()
        iml = [{"from":k,"import":tb.i_modules_info[k]} for k in iml]
        self.client.send({"type":"vpm_status","data":{"index":{"repo":self.vpm.get_repo_times(),"pkglist":res,"most_rated":mrt,"most_downloaded":mdn,"most_recent":mcr,"i_modules":iml}}})

    def send_vms(self):
        #remove binaries
        self.vms = tb.get_vms()
        vms = copy.deepcopy(self.vms)
        for vm in vms:
            del vms[vm]["bin"]
        self.client.send({"type":"vms","data":[v for k,v in vms.items()]})

    def get_feed(self,msg):
        self.client.answer(msg,"ok","")
        self.send_feed()

    def send_feed(self):
        if self.backend.connected:
            try:
                d = feedparser.parse("http://community.viperize.it/categories/releases/feed.rss")
                self.client.send({"type":"feed","data":d.entries})
            except:
                pass

    def check_devices(self):
        if self.hwlistener.fresh_boards:
            plgd,uplg = self.hwlistener.get_boards()
            tsbrd = {}
            for u,b in plgd.items():
                tsbrd[b.uid]={
                    "device":b.uid,
                    "board":b.board,
                    "name":b.desc,
                    "ram":b.ram,
                    "rom":b.flash,
                    "reset":b.manual_reset,
                    "port":b.port,
                    "disk":b.disk,
                    "restorable":hasattr(b,"restore"),
                    "viperizable": b.burn(None) if hasattr(b,"not_viperizable") else True,
                    "virtual":False
                    }
            for u,b in uplg.items():
                tsbrd[b.uid]={
                    "device":b.uid,
                    "board":b.board,
                    "name":b.desc,
                    "virtual":True
                    }
            self.devices = tsbrd
            self.send_devices()


###### UTILS

    def synch_with_remote(self):
        if self.backend.logged:
            prjs = self.backend.get_projects()
            found = False
            for proj in prjs:
                for uid,prj in self.projects.items():
                    if prj["name"]==proj["name"] and ("remoteId" not in prj or prj["remoteId"]==proj["remoteId"]):
                        prj["remoteId"]=proj["remoteId"]
                        #run git to synch
                        self.say("SYNC","Updating project "+prj["name"])
                        self.git.synch_cloud(ide.viper.viper_origin,self.dname(proj))
                        found = True
                        break
                if not found:
                    #new project
                    uid = self._new_project_dir(proj)
                    self.say("SYNC","Downloading remote project "+prj["name"])
                    self.git.new_repo(self.dname(proj))
                    self.git.add_remote(ide.viper.viper_origin,self.backend.get_project_repo_url(proj))
                    self.git.synch_cloud(ide.viper.viper_origin,self.dname(proj))
                    fill_from_path(proj)
        self.send_projects()



    def _get_board(self,dev):
        plgd,uplg = self.hwlistener.get_boards()
        if dev["device"] not in plgd and dev["device"] not in uplg:
            return None
        if dev["device"] in plgd:
            return plgd[dev["device"]]
        else:
            return uplg[dev["device"]]

    def _get_and_extract_archive(self,url,path):
        with tempfile.TemporaryFile(prefix="_import",dir=vconf.envdirs["tmp"]) as tf:
            c = pycurl.Curl()
            c.setopt(c.URL, url)
            c.setopt(c.WRITEDATA, tf)
            c.perform()
            c.close()
            tf.seek(0,0)
            zipf = zipfile.ZipFile(tf)
            zipf.extractall(path)

    def _get_repo(self,proj,name):
        for repo in proj["repos"]:
            if repo["name"]==name:
                return repo
        return None

    def _save_project(self,proj):
        for fname in proj["files"]:
            #print("TRYING TO SAVE:",fname)
            if fname.endswith(os.path.sep):
                #it's a dir
                try:
                    os.mkdir(self.fname(proj,fname))
                except:
                    pass
            else:
                with open(self.fname(proj,fname),"wb") as ff:
                    ff.write(base64.b64decode(bytes(proj["files"][fname],"utf-8")))
            proj["files"][fname]=""
        dname = self.dname(proj)
        for fl in self.find_files(dname,"*"):
            if os.path.relpath(fl,dname) not in proj["files"]:
                os.remove(fl)
        self.bcodes[proj["id"]]=None


    def _project_exists(self,name):
        for proj in self.projects:
            if self.projects[proj]["name"]==name:
                return self.projects[proj]
        return False

    def generate_main_py(self,proj,encoded = True):
        res = "#"*80
        res+="\n"
        res+= "# "+proj["name"]+"\n"
        res+= "#\n"
        res+= "# Created: "+str(datetime.datetime.now())+"\n"
        res+= "#\n"
        res+= "#"*80
        res+="\n\n"
        if not encoded:
            return res
        return self.data_encode(res)
        #base64.b64encode(res.encode("utf-8")).decode("utf-8")

    def generate_library_main_py(self,proj):
        res ="# use this file for testing the library only,\n"
        res+="# it will not be included when the library is released\n\n"
        if proj["local"]:
            res+="from local."+proj["mod_name"]+" import "+proj["mod_name"]
        else:
            res+="from community."+proj["namespace"]+"."+proj["mod_name"]+" import "+proj["mod_name"]
        res+="\n"
        return self.data_encode(self.generate_main_py(proj,False) + res)

    def generate_project_md(self,proj):
        res = proj["name"]+"\n"
        res+= ("="*len(proj["name"]))+"\n\n"
        res+= proj["desc"]+"\n"
        return self.data_encode(res)
        #base64.b64encode(res.encode("utf-8")).decode("utf-8")

    def generate_library_docs_json(self,proj):
        res = {'version': '', 'title': proj['name'], 'copyright': ''}
        return self.data_encode(json.dumps(res,sort_keys=True,indent=4))

    def generate_library_docs_rst(self,proj):
        res = ".. _" + proj['mod_name']
        res+= "\n\n"
        res+= "*"*len(proj["name"])
        res+= "\n"+proj["name"]+"\n"
        res+= "*"*len(proj["name"])
        res+= "\n\n"
        res+= ".. include:: __toc.rst\n\n"
        return self.data_encode(res)

    def store(self):
        self.db.store_projects(self.projects)
        self.db.store_devices(self.dev_props)

    def dname(self,proj):
        return os.path.join(self.projdir,proj["id"])

    def fname(self,proj,fname):
        return os.path.join(self.projdir,proj["id"],fname)

    def stripfname(self,proj,fname):
        prf = os.path.commonprefix([fname,self.dname(proj)])
        ret = fname[len(prf):]
        if ret.startswith(os.path.sep):
            ret = ret[1:]
        return ret

    def data_encode(self,data):
        if type(data)==str:
            data = data.encode("utf-8")
        return base64.b64encode(data).decode("utf-8")

    def fill_project(self,proj):
        for fname in proj["files"]:
            logger.info("filling project with file %s",fname)
            if not os.path.isdir(self.fname(proj,fname)):
                with open(self.fname(proj,fname),"rb") as ff:
                    proj["files"][fname]=self.data_encode(ff.read())
                    #base64.b64encode(ff.read()).decode("utf-8")

    def unfill_project(self,proj):
        for fname in proj["files"]:
                proj["files"][fname]=""

    def find_files(self,directory, pattern):
        for root, dirs, files in os.walk(directory):
            #ignore paths containing . (.git and subdirectories for example)
            #print("ROOT:",root.replace(directory,""),os.path.split(root)[0])
            if "." in root.replace(directory,""):
                continue
            for basename in files:
                if fnmatch.fnmatch(basename, pattern) and not basename.startswith("."):
                    filename = os.path.join(root, basename)
                    yield filename

    def fill_from_path(self,proj):
        proj["files"]={}
        for fl in self.find_files(self.dname(proj),"*"):
            #print("FILLING WITH:",fl)
            proj["files"][os.path.relpath(fl,self.dname(proj))]=""




class FakeConsole(ViperThread):
    def __init__(self,name):
        super().__init__()
        self.name = name
        self.running = False
        self.ctrl = ide.viper.getController()
        self.start()
    def run(self):
        while self.running:
            self.ctrl.say_console(self.name,str(uuid.uuid4())+"\n")
            time.sleep(1)
    def start(self):
        self._start(self.run,"Console ["+self.name+"]")

import serial
import queue

class ViperConsole(ViperThread):
    def __init__(self,name,board):
        super().__init__()
        self.name = name
        self.running = False
        self.ctrl = ide.viper.getController()
        self.board = board
        self.has_close = threading.Event()
        self.has_close.clear()
        self.iqueue = queue.Queue()
        self.start()
    def send(self,x):
        try:
            if self.ser:
                self.ser.write(bytes(x,"utf-8"))
                self.ser.flush()
        except:
            pass
    def run(self):
        #print("Connecting to",self.board.addr)

        try:
            if self.board.addr:
                myport = self.board.addr
                sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
                sock.settimeout(5)
                sock.connect(self.board.addr)
                sock.settimeout(0.3)
                self.ser = tb.AStream(sock)
            else:
                myport = self.board.port
                self.ser= serial.Serial(self.board.port,self.board.baud,timeout=2)
                self.ser.timeout=0.3
                self.ser = tb.AStream(self.ser)
        except Exception as ee:
            self.ser = None
            print(ee)
            try:
                ide.viper.getController().notification("Port "+str(myport)+" can't be opened! Please reset the board")
                ide.viper.getController().say_close_console(self.name)
                del ide.viper.getController().consoles[self.name]
            except Exception as e:
                logger.info("Exception deleting Console %s on %s is %s",self.name,self.board.port,str(e))

        if self.ser:
            try:
                while self.running:
                    msg = self.ser.readline()
                    if msg:
                        self.ctrl.say_console(self.name,msg)
                    #print(".")
            except:
                #print("#####")
                try:
                    ide.viper.getController().say_close_console(self.name)
                    del ide.viper.getController().consoles[self.name]
                except Exception as e:
                    logger.info("Exception deleting Console %s on %s is %s",self.name,self.board.port,str(e))
            finally:
                try:
                    self.ser.close()
                except Exception as e:
                    logger.info("Exception closing Console %s on %s is %s",self.name,self.board.port,str(e))
                logger.info("Closed Console %s on %s",self.name,self.board.port)
        self.has_close.set()
    def start(self):
        self._start(self.run,"Console ["+self.name+"]")
        logger.info("Started Console %s on %s",self.name,self.board.port)
