import sys
import re
import sys
import glob
import os.path
import importlib
import threading
import time
import platform
import subprocess

import ide.viper
from ide.viperthreads import ViperThread
import viper.viperlogger
import viper.viperconf as vconf
import viper.toolbelt as tb

logger = viper.viperlogger.logger(__name__,"ide")



### LOAD BOARD MODULES
#boardfiles = glob.glob(os.path.join(os.path.dirname(os.path.realpath(__file__)),"boards","*.py"))
#sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)),"boards"))

tb.reload_boards()

if vconf.platform=="linux":
    import pyudev
elif vconf.platform=="windows":
    from ide.wmi import wmi
    import win32com.client
    import pythoncom
elif vconf.platform=="mac":
    import ide.macusb as macusb


class ViperListener(ViperThread):
    def __init__(self):
        super().__init__()
        self.boards_plugged={}
        self.fresh_boards = True
        self.devices = {}
        self.b_lock = threading.Lock()
        if vconf.platform=="linux":
            self.runfn=self.run_linux
            self.setupfn=self.setup_linux
            self.stopfn=self.stop_linux
        elif vconf.platform=="windows":
            self.runfn=self.run_win
            self.setupfn=self.setup_win
            self.stopfn=self.stop_win
        elif vconf.platform=="mac":
            self.runfn=self.run_mac
            self.setupfn=self.setup_mac
            self.stopfn=self.stop_mac
        # self.load_boards()

    def load_boards(self):
        tb.reload_boards()
        self.vendors = {}
        self.models = set()
        self.board_by_vendor = {}
        for board in tb.boards:
            if hasattr(board,"ids_vendor"):
                for x in board.ids_vendor:
                    if x not in self.vendors:
                        self.vendors[x]=set()
                    self.vendors[x].update(board.ids_vendor[x])
                    if x not in self.board_by_vendor:
                        self.board_by_vendor[x]=set()
                    self.board_by_vendor[x].add(board)
        print(self.vendors)

    def get_boards(self):
        self.b_lock.acquire()
        self.boards_plugged=[]
        self.boards_unplugged=[]
        for brd in tb.boards:
            bb = brd.discover(self.devices)
            if bb:
                self.boards_plugged.extend(bb)
            else:
                if not hasattr(brd,"not_virtual"):
                    bbb = brd({"port":"X","disk":"X","uid":brd.board+"_"+brd.__name__+"_virtual_","vendor":"0000","model":"0000","devstring":"X"})
                    self.boards_unplugged.append(bbb)
        plgd = {x.uid:x for x in self.boards_plugged}
        uplg = {x.uid:x for x in self.boards_unplugged}
        self.fresh_boards=False
        self.b_lock.release()
        return (plgd,uplg)
    def refresh(self):
        self.b_lock.acquire()
        tb.reload_boards()
        self.b_lock.release()
    def start(self):
        self._start(self.run,"Device Listener")
    def run(self):
        self.setupfn()
        while self.running:
            self.runfn()
        self.stopfn
    def _make_uid(self,vid,pid,sid):
        return vid.upper()+"_"+pid.upper()+"_"+sid.upper()

######################################## LINUX
    def setup_linux(self):
        self.keys = frozenset(["ID_SERIAL","ID_SERIAL_SHORT","ID_VENDOR_ID","ID_MODEL_ID","SUBSYSTEM","DEVNAME"])
        self.context = pyudev.Context()
        logger.info("Listing all devices")
        for device in self.context.list_devices():
            if device["SUBSYSTEM"] in ("block","tty","usb"):
                uid, dev = self.check_dev_linux(device)
                if uid:
                    self.b_lock.acquire()
                    self.devices[uid]=dev
                    self.fresh_boards=True
                    self.b_lock.release()

        logger.info("Creating UDEV Monitor")
        self.monitor = pyudev.Monitor.from_netlink(self.context)
        self.monitor.filter_by("tty")
        self.monitor.filter_by("block")
        self.monitor.filter_by("usb")
        self.monitor.start()
    def stop_linux(self):
        pass
    def check_dev_linux(self,dev,action="add"):
        uid=""
        device = {}
        if self.keys.issubset(dev):
            uid = self._make_uid(dev["ID_VENDOR_ID"],dev["ID_MODEL_ID"],dev["ID_SERIAL_SHORT"])
            if uid in self.devices:
                device=self.devices[uid]
            else:
                device={}
            if device and action=="remove":
                return (uid,device)
            device["serial"]=dev["ID_SERIAL_SHORT"]
            device["devstring"]=dev["ID_SERIAL"]
            device["vendor"]=dev["ID_VENDOR_ID"].upper()
            device["model"]=dev["ID_MODEL_ID"].upper()
            device["uid"] = uid
            if dev["SUBSYSTEM"]=="tty":
                device["port"]=  dev["DEVNAME"]
            elif dev["SUBSYSTEM"]=="block":
                block = dev["DEVNAME"]
                # if device["vendor"] in self.vendors and device["model"] in self.vendors[device["vendor"]]:
                    #go on and check mount points
                    # pass
                # else:
                    # return ("",{})
                attempts=0
                while action=="add":
                    ret,out = tb.runcmd("mount","-v")
                    logger.info("Checking mount points for %s",uid)
                    device["disk"]=""
                    if not ret:
                        paths = out.split("\n")
                        for path in paths:
                            words = path.split()
                            if len(words)>=4:
                                #Mmm...what if there are spaces in the path name!?!
                                if words[0].startswith(block) and words[1]=="on" and words[3]=="type":
                                    if words[0]==block:
                                        device["disk"]=words[2]
                                        logger.info("FOUND!!! %s = %s",block,device["disk"])
                                        break
                                    else:
                                        #since it startswith, it has partitions..ignore it and go to next device
                                        return ("",{})
                    if device["disk"]:
                        break
                    else:
                        time.sleep(1)
                        attempts+=1
                        if attempts>10:
                            return ("",{})
        return(uid,device)
    def run_linux(self):
        dev = self.monitor.poll(5)
        if dev:
            uid,device = self.check_dev_linux(dev,dev.action)
            if uid:
                if dev.action=="remove" and uid in self.devices:
                    del self.devices[uid]
                elif dev.action=="add":
                    self.devices[uid]=device
                logger.info("DEVICE (%s): %s",dev.action,device["uid"])
                self.b_lock.acquire()
                self.fresh_boards=True
                self.b_lock.release()

######################################## WINDOWS
    def setup_win(self):
        self.smth = re.compile("USB\\\\(?:[^\\\\]*&)*VID_([0-9a-fA-F]*)&PID_([0-9a-fA-F]*)(?:&[^\\\\]*)*\\\\([^\\\\]*)")
        self.hmth = re.compile('.*="(.*)"')
        self.devices = {}
        pythoncom.CoInitialize()
        self.wmic = win32com.client.GetObject ("winmgmts:")
        self.wmi = wmi.WMI()
    def stop_win(self):
        pythoncom.CoUninitialize()
    def check_dev_win(self,dev,action="add"):
        pass
    def _get_win_device_id(self,pnp):
        mth =self.smth.match(pnp)
        if mth:
            return (
                mth.group(1).upper(),
                mth.group(2).upper(),
                mth.group(3).upper(),
                self._make_uid(mth.group(1),mth.group(2),mth.group(3))
                )
        return (None,None,None,None)
    def _split_sid(self,sid):
        if "&" in sid:
            return sid.split("&")[1]
        return None
    def _get_hw_specs(self):
        pnps = {}
        curpnp = None
        curvid = None
        curpid = None
        for usb in self.wmi.InstancesOf ("Win32_USBControllerDevice"):
            mth = self.hmth.match(usb.Dependent)
            if mth:
                unescaped = bytes(mth.group(1),"utf-8").decode('unicode_escape')
                vid,pid,sid,uid = self._get_win_device_id(unescaped)
                if vid:
                    if "&" in sid:
                        #not a device spec, get instance id code
                        iid = self._split_sid(sid)
                        if curpnp and vid==curvid and pid==curpid:
                            pnps[curpnp].add(iid)
                    else:
                        if curpnp!=sid:
                            pnps[sid]=set()
                            curpnp=sid
                            curvid=vid
                            curpid=pid
        return pnps
    def _get_hw(self,sids=[],nouids=[]):
        pnps = {}
        curpnp=None
        for usb in self.wmi.InstancesOf ("Win32_PNPEntity"):
            #unescaped = bytes(usb.PNPDeviceID,"utf-8").decode('unicode_escape')
            vid,pid,sid,uid = self._get_win_device_id(usb.PNPDeviceID)
            if uid and uid not in nouids and sid in sids:
                #print("HW==>",uid)
                if curpnp!=sid:
                    pnps[sid]=usb
                    curpnp=sid
        return pnps
    def _get_drive_letter_from_id(self,sid):
        for physical_disk in self.wmi.Win32_DiskDrive (InterfaceType="USB"):
            for partition in physical_disk.associators ("Win32_DiskDriveToDiskPartition"):
                for logical_disk in partition.associators ("Win32_LogicalDiskToPartition"):
                    if sid in physical_disk.PNPDeviceID:
                        return logical_disk.DeviceID
        return ""
    def run_win(self):
        self.cdevs={}
        dfnd = set()
        time.sleep(1)
        pnps = self._get_hw_specs()
        serdev =[]
        #print("PNPS=>",pnps)
        for x in self.wmi.Win32_SerialPort():
            vid,pid,sid,uid = self._get_win_device_id(x.PNPDeviceID)
            if sid:
                ddisk=""
                if sid in pnps:
                    #it's a non composite usb serial device
                    pass
                else:
                    #it's a composite device, hack a bit
                    ssid = self._split_sid(sid)
                    uid =None
                    ddisk = ""
                    for k,v in pnps.items():
                        if ssid in v:
                            uid = self._make_uid(vid,pid,k)
                            ddisk = self._get_drive_letter_from_id(k)
                            break
                if uid:
                    dev = {
                        "uid":uid,
                        "vendor":vid,
                        "model":pid,
                        "serial":sid,
                        "port":x.DeviceID,
                        "devstring":x.Description,
                        "disk":ddisk
                    }
                    dfnd.add(uid)
                    self.cdevs[uid]=dev
                    serdev.append(uid)
        usbdevs = self._get_hw(pnps.keys(),serdev)
        #print("SERDEV",serdev)
        for iid,usbdev in usbdevs.items():
            #print(iid,"==>",usbdev.PNPDeviceID)
            vid,pid,sid,uid = self._get_win_device_id(usbdev.PNPDeviceID)
            if uid:
                dev = {
                    "uid":uid,
                    "vendor":vid,
                    "model":pid,
                    "serial":sid,
                    "port":"NULL",
                    "devstring":usbdev.Name,
                    "disk":""
                }
                dfnd.add(uid)
                self.cdevs[uid]=dev
        #print(self.cdevs)
        ddnd = set()
        for k in self.devices:
            ddnd.add(k)
        if dfnd>=ddnd and dfnd<=ddnd:
            return
        self.b_lock.acquire()
        self.devices = self.cdevs
        self.fresh_boards=True
        self.b_lock.release()

######################################## MAC
    def setup_mac(self):
        self.macusb = macusb.MacUsb()
    def stop_mac(self):
        pass
    def check_dev_mac(self,dev,action="add"):
        pass
    def _get_mount_point_mac(self,disk):
        try:
            out = subprocess.check_output("diskutil info "+disk,universal_newlines=True,shell=True)
            lines = out.split("\n")
            for line in lines:
                if "Mount Point:" in line:
                    return line.replace("Mount Point:","").strip()
            return disk
        except:
            return disk
    def ioreg(self):
        try:
            return subprocess.check_output("ioreg -x -r -l -c "+self.iousbstring,universal_newlines=True,shell=True)
        except:
            return ""
    def parse(self):
        nodes = self.macusb.get_nodes()
        devs = set()
        nodes_ok = set()
        props = {
            "idVendor":"vendor",
            "idProduct":"model",
            "USB Serial Number":"serial",
            "USB Product Name":"serial_id",
            "IODialinDevice":"port",
            "BSD Name":"disk"
        }
        devices = []
        for iid,node in nodes.items():
            if node.id in nodes_ok:
                #print(node.name,"excluded")
                continue
            nodes_ok.add(node.id)
            vendor = node.has_vendor()
            if vendor and vendor in self.vendors and node.cls==self.macusb.iousbstring:
                dev = {"devstring":node.name}
                candidates = [node]+[nodes[x] for x in node.descendants]
                for k,v in props.items():
                    for dnode in candidates:
                        if k in dnode.properties():
                            dev[v]=dnode.get_property(k).strip('"')
                            if v=="disk":
                                dev[v]=self._get_mount_point_mac(dev[v])
                            break
                nodes_ok.update(node.descendants)
                dev["uid"]=self._make_uid(dev["vendor"],dev["model"],dev["serial"])
                devices.append(dev)
        devlist = {}
        for dev in devices:
            devlist[dev["uid"]]=dev
        return devlist

        # dev = {}
        # for k,node in nodes.items():

        #     if line.startswith("+"):
        #         #got a device
        #         if dev:
        #             devices.append(dev)
        #         dev = {"devstring":line[:line.find("<")]}
        #     elif "idVendor" in line:
        #         dev["vendor"]=line[line.find("0x")+2:].upper()
        #     elif "idProduct" in line:
        #         dev["model"]=line[line.find("0x")+2:].upper()
        #     elif "USB Serial Number" in line:
        #         dev["serial"]=line[line.find("= \"")+3:-1].upper()
        #     elif "USB Product Name" in line:
        #         dev["serial_id"]=line[line.find("= \"")+3:-1]
        #     elif "IODialinDevice" in line:
        #         dev["port"]=line[line.find("= \"")+3:-1]
        #     elif "BSD Name" in line:
        #         dev["disk"]=line[line.find("= \"")+3:-1]
        #         dev["disk"]=self._get_mount_point_mac(dev["disk"])
        # if dev and dev not in devices:
        #     devices.append(dev)

        # devlist = {}
        # for dev in devices:
        #     if "model" in dev and "vendor" in dev and "serial" in dev:
        #         vendor = dev["vendor"].strip()
        #         model = dev["model"].strip()
        #         if len(vendor)<4:
        #             vendor=("0"*(4-len(vendor)))+vendor
        #         if len(model)<4:
        #             model=("0"*(4-len(model)))+model
        #         dev["uid"]=self._make_uid(vendor,model,dev["serial"])
        #         devlist[dev["uid"]]=dev
        #         dev["vendor"]=vendor
        #         dev["model"]=model
        # return devlist

    def run_mac(self):
        time.sleep(1)
        devs = self.parse()
        #print(devs)
        deepequal = True
        if devs.keys()==self.devices.keys():
            for k,v in devs.items():
                if v.keys()!=self.devices[k].keys():
                    deepequal=False
                    break
        else:
            deepequal=False
        if not deepequal:
            self.b_lock.acquire()
            self.devices = devs
            self.fresh_boards=True
            self.b_lock.release()
