import subprocess
import os.path
from urllib.parse import urlparse
import re
import ide.viper
import viper.toolbelt as tb
import viper.viperconf as vconf

class GitException(Exception):
    def __init__(self,msg):
        super().__init__()
        self.msg=msg
    def __str__(self):
        return self.msg

class ViperGit():
    def __init__(self):
        ##create git credentials foo
        with open(vconf.gitfoo,"wt") as gitf:
            if vconf.platform.startswith("win"):
                gitf.write("echo \"username=nouser\"\r\necho \"password=nopass\"\r\n")
            else:
                gitf.write("echo \"username=nouser\"\necho \"password=nopass\"\n")
        os.chmod(vconf.gitfoo,0o755)

    def origin(self,repo,user,password):
        o = urlparse(repo)
        if user and password:
            crd = user+":"+password+"@"
        elif user:
            crd = user+"@"
        else:
            crd=""
        if not o.scheme:
            scheme="file://"
        else:
            scheme=o.scheme+"://"
        return scheme+crd+o.netloc+o.path
    def prepare_commit(self,repo):
        res,out = tb.run("git", "add","-A",cwd=repo)
        if res:
            raise GitException(out)
        return out
    def status(self,repo):
        res,out = tb.run("git", "status","--porcelain",cwd=repo)
        if res:
            raise GitException(out)
        return out
    def commit(self,repo,msg):
        out = self.status(repo)
        if out: #there is something to commit
            res,out = tb.run("git", "commit","-m","\""+msg+"\"",cwd=repo)
            if res:
                raise GitException(out)
        return out
    def do_commit(self,repo,msg):
        self.prepare_commit(repo)
        return self.commit(repo,msg)
    def push(self,repo,origin,branch="master"):
        res,out = tb.run("git", "push","-u",origin,branch,cwd=repo,datain="nouser\nnopass\n")
        if res:
            raise GitException(out)
        return out
    def first_push(self,repo,origin,branch="master"):
        res,out = tb.run("git", "push","-u",origin,branch,cwd=repo,datain="nouser\nnopass\n")
        if res:
            raise GitException(out)
        return out
    def clone(self,repo,origin,origin_name="origin"):
        res,out = tb.run("git","-c","http.sslVerify=false", "-c", "credential.helper="+vconf.gitfoo.replace("\\","/"),"clone","-o",origin_name,origin,".",cwd=repo,datain="nouser\nnopass\n")
        if res:
            raise GitException(out)
        return self.config_repo(repo)
    def log(self,repo):
        res,out = tb.run("git", "push",origin,"master",cwd=repo,datain="nouser\nnopass\n")
        lines = out.split("\n")
        commits = []
        curcommit = None
        c0 = re.compile("commit (.*)")
        c1 = re.compile("Author: (.*)")
        c2 = re.compile("Date: (.*)")
        c3 = re.compile(" (.*)")
        for line in lines:
            mth = c0.match(line)
            if mth:
                curcommit={"hash":c0.group(1)}
                commits.append(curcommit)
            else:
                mth = c1.match(line)
                if mth and curcommit:
                    curcommit["author"]=c1.group(1)
                else:
                    mth = c2.match(line)
                    if mth and curcommit:
                        curcommit["date"]=c2.group(1)
                    else:
                        mth = c3.match(line)
                        if mth and curcommit:
                            txt = c3.group(1).strip()
                            if "text" not in curcommit:
                                curcommit["text"]=""
                            if txt:
                                curcommit["text"]+=txt
        if res:
            raise GitException(out)
        return commits
    def dirflags(self,dname):
        return "--git-dir="+os.path.abspath(os.path.join(dname,".git"))+" --work-tree="+os.path.abspath(dname)
   
    def reset_to_master(self,repo,origin):
        res,out = tb.run("git", "reset", "--hard",origin+"/master",cwd=repo)
        if res:
            raise GitException(out)
        return out
   
    def config_repo(self,repo):
        name = ide.viper.getBackend().getUser() or "Viper"
        email = ide.viper.getBackend().getEmail() or "nomail@viperize.it"
        res,out = tb.run("git","config","http.sslVerify","false",cwd=repo)
        if res:
            raise GitException(out)
        res,out = tb.run("git","config","user.name",'"'+name+'"',cwd=repo)
        if res:
            raise GitException(out)
        res,out = tb.run("git","config", "user.email",'"'+email+'"',cwd=repo)
        if res:
            raise GitException(out)
        return out

    def new_repo(self,path):
        res,out = tb.run("git","init",cwd=path)
        if res:
            raise GitException(out)
        return self.config_repo(path)

    def add_remote(self,remote,url,path):
        #print(tools.conf.gitfoo.replace("\\","/"))
        res,out = tb.run("git","config","credential.helper",vconf.gitfoo.replace("\\","/"),cwd=path)
        res,out = tb.run("git","remote","add",remote,url,cwd=path)
        res,out = tb.run("git","remote","set-url",remote,url,cwd=path)
        if res:
            raise GitException(out)
        return out

    def remote_update(self,remote,repo):
        res,out = tb.run("git","remote","update",remote,cwd=repo,datain="nouser\nnopass\n")
        if res:
            raise GitException(out)
        res,out = tb.run("git","log","HEAD.."+remote+"/master","--oneline",cwd=repo,datain="nouser\nnopass\n")
        if res:
            raise GitException(out)
        if out:
            return True
        return False

    def merge(self,remote,repo):
        res,out = tb.run("git","merge","-X","theirs",remote+"/master", "-m=\"Automatic_Merge\"",cwd=repo,datain="nouser\nnopass\n")
        if res:
            raise GitException(out)
        return out


    def check_status(self,repo,origin):
        res,out = tb.run("git","remote","update",origin,cwd=repo,datain="nouser\nnopass\n")
        if res:
            raise GitException(out)
        res,out = tb.run("git","status",cwd=repo)
        if res:
            raise GitException(out)
        if "branch is ahead" in out:
            return "ahead"
        elif "branch is behind" in out:
            return "behind"
        elif "have diverged" in out:
            return "diverged"
        elif "branch is up-to-date" in out or "nothing to commit" in out:
            return "uptodate"
        raise GitException(out)


    def push_whatever(self,repo,origin,branch="master",tempbranch="temp"):
        status = self.check_status(repo,origin)

        if status=="diverged" or status=="behind":
            #now it comes the fun
            res,out = tb.run("git","branch","-D",tempbranch,cwd=repo)
            res,out = tb.run("git","checkout","-b",tempbranch,cwd=repo)
            if res:
                raise GitException(out)
            self.do_commit(repo,"branch_commit")
            res,out = tb.run("git","checkout","master",cwd=repo)
            if res:
                raise GitException(out)
            res,out = tb.run("git","reset", "--hard",origin+"/"+branch,cwd=repo)
            if res:
                raise GitException(out)
            #enter the magic
            res,out = tb.run("git","merge",tempbranch,"-X", "theirs", "-m=\"branch_merge\"",cwd=repo,datain="nouser\nnopass\n")
            if res:
                raise GitException(out)
        
        #push, at last
        res,out = tb.run("git", "push","-u",origin,branch,cwd=repo,datain="nouser\nnopass\n")
        if res:
            raise GitException(out)
        return out

# git branch -D work
# git checkout -b work
# git add .
# git commit -m "branch_commit"
# git checkout master
# git reset --hard viper_repo_cloud/master
# git merge work -X theirs -m "branch_merge"
# git push viper_repo_cloud master


    def synch_cloud(self, remote, repo):
        res,out = tb.run("git","fetch","remote",cwd=repo)
        res,out = tb.run("git","log","HEAD.."+remote+"/master","--oneline",cwd=repo)
        if out:
            #need to merge
            tb.run("git","merge","-X","theirs",remote+"/master",cwd=repo)
        else:
            res,out = tb.run("git","status","-sb",cwd=repo)
            if "ahead" in out:
                #need to commit & sync
                self.prepare_commit(repo)
                self.commit(repo,"sync")
                self.push(repo,remote)
            else:
                #nothing to do
                pass

